"""HuggingFace RAG Trainer"""

from typing import TYPE_CHECKING, Any, Callable, cast

from typing_extensions import assert_never

from fed_rag.base.trainer import BaseTrainer
from fed_rag.base.trainer_manager import BaseRAGTrainerManager, RAGTrainMode
from fed_rag.decorators import federate
from fed_rag.exceptions import (
    MissingExtraError,
    UnspecifiedGeneratorTrainer,
    UnspecifiedRetrieverTrainer,
)
from fed_rag.types.results import TestResult, TrainResult

try:
    from datasets import Dataset

    _has_huggingface = True
except ModuleNotFoundError:
    _has_huggingface = False


if TYPE_CHECKING:  # pragma: no cover
    from datasets import Dataset
    from sentence_transformers import SentenceTransformer

    from fed_rag.fl_tasks.huggingface import HFModelType, HuggingFaceFLTask


class HuggingFaceRAGTrainerManager(BaseRAGTrainerManager):
    """HuggingFace RAG Trainer Manager"""

    def __init__(
        self,
        mode: RAGTrainMode,
        retriever_trainer: BaseTrainer | None = None,
        generator_trainer: BaseTrainer | None = None,
        **kwargs: Any,
    ):
        if not _has_huggingface:
            msg = (
                f"`{self.__class__.__name__}` requires `huggingface` extra to be installed. "
                "To fix please run `pip install fed-rag[huggingface]`."
            )
            raise MissingExtraError(msg)
        super().__init__(
            mode=mode,
            retriever_trainer=retriever_trainer,
            generator_trainer=generator_trainer,
            **kwargs,
        )

    def _prepare_generator_for_training(self, **kwargs: Any) -> None:
        self.generator_trainer.model.train()

        # freeze generator
        if self.retriever_trainer:
            self.retriever_trainer.model.eval()

    def _prepare_retriever_for_training(
        self, freeze_context_encoder: bool = True, **kwargs: Any
    ) -> None:
        self.retriever_trainer.model.train()

        # freeze generator
        if self.generator_trainer:
            self.generator_trainer.model.eval()

    def _train_retriever(self, **kwargs: Any) -> TrainResult:
        if self.retriever_trainer:
            self._prepare_retriever_for_training()
            return self.retriever_trainer.train()
        else:
            raise UnspecifiedRetrieverTrainer(
                "Attempted to perform retriever trainer with an unspecified trainer."
            )

    def _train_generator(self, **kwargs: Any) -> TrainResult:
        if self.generator_trainer:
            self._prepare_generator_for_training()
            return self.generator_trainer.train()
        else:
            raise UnspecifiedGeneratorTrainer(
                "Attempted to perform generator trainer with an unspecified trainer."
            )

    def train(self, **kwargs: Any) -> TrainResult:
        if self.mode == "retriever":
            return self._train_retriever()
        elif self.mode == "generator":
            return self._train_generator()
        else:
            assert_never(self.mode)  # pragma: no cover

    def _get_federated_trainer(self) -> tuple[Callable, "HFModelType"]:
        if self.mode == "retriever":
            if self.retriever_trainer is None:
                raise UnspecifiedRetrieverTrainer(
                    "Cannot federate an unspecified retriever trainer."
                )
            retriever_train_fn = self.retriever_trainer.train
            retriever_module = self.retriever_trainer.model
            retriever_module = cast("SentenceTransformer", retriever_module)

            # Create a standalone function for federation
            def train_wrapper(
                model: "HFModelType",
                train_dataset: "Dataset",
                val_dataset: "Dataset",
            ) -> TrainResult:
                _ = retriever_train_fn()
                return TrainResult(loss=0)

            return (
                federate.trainer.huggingface(train_wrapper),
                retriever_module,
            )

        elif self.mode == "generator":
            if self.generator_trainer is None:
                raise UnspecifiedGeneratorTrainer(
                    "Cannot federate an unspecified generator trainer."
                )
            generator_train_fn = self.generator_trainer.train
            generator_module = self.generator_trainer.model

            # Create a standalone function for federation
            def train_wrapper(
                model: "HFModelType",  # TODO: handle union types in inspector
                train_dataset: "Dataset",
                val_dataset: "Dataset",
            ) -> TrainResult:
                _ = generator_train_fn()
                # TODO get loss from out
                return TrainResult(loss=0)

            return (
                federate.trainer.huggingface(train_wrapper),
                generator_module,
            )
        else:
            assert_never(self.mode)  # pragma: no cover

    def get_federated_task(self) -> "HuggingFaceFLTask":
        from fed_rag.fl_tasks.huggingface import HuggingFaceFLTask

        federated_trainer, _module = self._get_federated_trainer()

        # TODO: add logic for getting evaluator/tester and then federate it as well
        # federated_tester = self.get_federated_tester(tester_decorator)
        # For now, using a simple placeholder test function
        def test_fn(
            model: "HFModelType", eval_dataset: "Dataset"
        ) -> TestResult:
            # Implement simple testing or return a placeholder
            return TestResult(loss=0.42, metrics={})  # pragma: no cover

        federated_tester = federate.tester.huggingface(test_fn)

        return HuggingFaceFLTask.from_trainer_and_tester(
            trainer=federated_trainer,
            tester=federated_tester,
        )



"""PyTorch Tester Inspector"""

import inspect
from typing import Any, Callable

from fed_rag.exceptions import (
    InvalidReturnType,
    MissingDataParam,
    MissingNetParam,
)
from fed_rag.inspectors.common import TesterSignatureSpec
from fed_rag.inspectors.huggingface.utils import get_type_name
from fed_rag.types import TestResult


def inspect_tester_signature(fn: Callable) -> TesterSignatureSpec:
    sig = inspect.signature(fn)

    # validate return type
    return_type = sig.return_annotation
    if (return_type is Any) or not issubclass(return_type, TestResult):
        msg = "Tester should return a fed_rag.types.TestResult or a subclass of it."
        raise InvalidReturnType(msg)

    # inspect fn params
    extra_tester_kwargs = []
    net_param = None
    test_data_param = None
    net_parameter_class_name = None

    for name, t in sig.parameters.items():
        if name in ("self", "cls"):
            continue

        if type_name := get_type_name(t):
            if (
                type_name
                in [
                    "PreTrainedModel",
                    "SentenceTransformer",
                    "PeftModel",
                    "HFModelType",
                ]
                and net_param is None
            ):
                net_param = name
                net_parameter_class_name = type_name
                continue

            if type_name == "Dataset" and test_data_param is None:
                test_data_param = name
                continue

        extra_tester_kwargs.append(name)

    if net_param is None:
        msg = (
            "Inspection failed to find a model param. "
            "For HuggingFace this param must have type `PreTrainedModel` or `SentenceTransformers`."
        )
        raise MissingNetParam(msg)

    if test_data_param is None:
        msg = (
            "Inspection failed to find a data param for a test dataset."
            "For HuggingFace these params must be of type `datasets.Dataset`"
        )
        raise MissingDataParam(msg)

    spec = TesterSignatureSpec(
        net_parameter=net_param,
        test_data_param=test_data_param,
        extra_test_kwargs=extra_tester_kwargs,
        net_parameter_class_name=net_parameter_class_name,
    )
    return spec



import re
import sys
from contextlib import nullcontext as does_not_raise
from unittest.mock import MagicMock, patch

import pytest

from fed_rag.exceptions import (
    InvalidDistanceError,
    KnowledgeStoreError,
    KnowledgeStoreNotFoundError,
    KnowledgeStoreWarning,
    LoadNodeError,
    MissingExtraError,
)
from fed_rag.knowledge_stores.qdrant.sync import (
    QdrantKnowledgeStore,
    _convert_knowledge_node_to_qdrant_point,
    _convert_scored_point_to_knowledge_node_and_score_tuple,
)
from fed_rag.types.knowledge_node import KnowledgeNode


def test_init() -> None:
    knowledge_store = QdrantKnowledgeStore(
        collection_name="test collection", load_nodes_kwargs={"parallel": 4}
    )

    assert isinstance(knowledge_store, QdrantKnowledgeStore)
    assert knowledge_store.load_nodes_kwargs == {"parallel": 4}


def test_init_raises_error_if_qdrant_extra_is_missing_parent_import() -> None:
    modules = {"qdrant_client": None}
    module_to_import = "fed_rag.knowledge_stores.qdrant"

    if module_to_import in sys.modules:
        original_module = sys.modules.pop(module_to_import)

    with patch.dict("sys.modules", modules):
        msg = (
            "Qdrant knowledge stores require the qdrant-client to be installed. "
            "To fix please run `pip install fed-rag[qdrant]`."
        )
        with pytest.raises(
            MissingExtraError,
            match=re.escape(msg),
        ):
            from fed_rag.knowledge_stores.qdrant import QdrantKnowledgeStore

            QdrantKnowledgeStore()

    # restore module so to not affect other tests
    sys.modules[module_to_import] = original_module


def test_init_raises_error_if_qdrant_extra_is_missing() -> None:
    modules = {"qdrant_client": None}
    module_to_import = "fed_rag.knowledge_stores.qdrant.sync"

    if module_to_import in sys.modules:
        original_module = sys.modules.pop(module_to_import)

    with patch.dict("sys.modules", modules):
        msg = (
            "Qdrant knowledge stores require the qdrant-client to be installed. "
            "To fix please run `pip install fed-rag[qdrant]`."
        )
        with pytest.raises(
            MissingExtraError,
            match=re.escape(msg),
        ):
            from fed_rag.knowledge_stores.qdrant.sync import (
                QdrantKnowledgeStore,
            )

            QdrantKnowledgeStore()

    # restore module so to not affect other tests
    sys.modules[module_to_import] = original_module


@patch("qdrant_client.QdrantClient")
def test_get_qdrant_client(mock_qdrant_client_class: MagicMock) -> None:
    knowledge_store = QdrantKnowledgeStore(
        collection_name="test collection",
    )

    # act
    with knowledge_store.get_client() as _client:
        pass

    mock_qdrant_client_class.assert_called_once_with(
        host="localhost",
        port=6333,
        grpc_port=6334,
        api_key=None,
        https=False,
        timeout=None,
    )


@patch("qdrant_client.QdrantClient")
def test_get_qdrant_client_ssl(mock_qdrant_client_class: MagicMock) -> None:
    knowledge_store = QdrantKnowledgeStore(
        collection_name="test collection", ssl=True
    )

    # act
    with knowledge_store.get_client() as _client:
        pass

    mock_qdrant_client_class.assert_called_once_with(
        host="localhost",
        port=6333,
        grpc_port=6334,
        api_key=None,
        https=False,
        timeout=None,
    )


@patch("qdrant_client.QdrantClient")
def test_get_qdrant_client_error_at_teardown_throws_warning(
    mock_qdrant_client_class: MagicMock,
) -> None:
    knowledge_store = QdrantKnowledgeStore(
        collection_name="test collection",
    )
    mock_instance = MagicMock()
    mock_qdrant_client_class.return_value = mock_instance
    mock_instance.close.side_effect = RuntimeError("mock error from qdrant")

    # act
    with pytest.warns(
        KnowledgeStoreWarning,
        match="Unable to close client: mock error from qdrant",
    ):
        with knowledge_store.get_client() as _client:
            pass

    mock_qdrant_client_class.assert_called_once_with(
        host="localhost",
        port=6333,
        grpc_port=6334,
        api_key=None,
        https=False,
        timeout=None,
    )
    mock_instance.close.assert_called_once()


@patch("qdrant_client.QdrantClient")
def test_load_node(mock_qdrant_client_class: MagicMock) -> None:
    mock_client = MagicMock()
    mock_qdrant_client_class.return_value = mock_client
    knowledge_store = QdrantKnowledgeStore(
        collection_name="test collection",
    )
    node = KnowledgeNode(
        node_id="1",
        embedding=[1, 1, 1],
        node_type="text",
        text_content="mock node",
    )

    # act
    knowledge_store.load_node(node)

    mock_client.collection_exists.assert_called_once_with("test collection")
    mock_client.upsert.assert_called_once_with(
        collection_name="test collection",
        points=[_convert_knowledge_node_to_qdrant_point(node)],
    )


@patch("qdrant_client.QdrantClient")
def test_load_node_raises_error(mock_qdrant_client_class: MagicMock) -> None:
    mock_client = MagicMock()
    mock_qdrant_client_class.return_value = mock_client
    knowledge_store = QdrantKnowledgeStore(
        collection_name="test collection",
    )
    node = KnowledgeNode(
        node_id="1",
        embedding=[1, 1, 1],
        node_type="text",
        text_content="mock node",
    )

    mock_client.upsert.side_effect = RuntimeError("mock error from qdrant")

    with pytest.raises(
        LoadNodeError,
        match="Failed to load node 1 into collection 'test collection': mock error from qdrant",
    ):
        knowledge_store.load_node(node)


@patch("qdrant_client.QdrantClient")
def test_load_nodes(mock_qdrant_client_class: MagicMock) -> None:
    mock_client = MagicMock()
    mock_qdrant_client_class.return_value = mock_client
    knowledge_store = QdrantKnowledgeStore(
        collection_name="test collection", load_nodes_kwargs={"parallel": 4}
    )
    nodes = [
        KnowledgeNode(
            node_id="1",
            embedding=[1, 1, 1],
            node_type="text",
            text_content="mock node",
        ),
        KnowledgeNode(
            node_id="2",
            embedding=[2, 2, 2],
            node_type="text",
            text_content="mock node",
        ),
    ]

    # act
    knowledge_store.load_nodes(nodes)

    mock_client.collection_exists.assert_called_once_with("test collection")
    mock_client.upload_points.assert_called_once_with(
        collection_name="test collection",
        points=[_convert_knowledge_node_to_qdrant_point(n) for n in nodes],
        parallel=4,
    )

    with does_not_raise():
        knowledge_store.load_nodes([])  # a no-op


@patch("qdrant_client.QdrantClient")
def test_load_nodes_raises_error(mock_qdrant_client_class: MagicMock) -> None:
    mock_client = MagicMock()
    mock_qdrant_client_class.return_value = mock_client
    knowledge_store = QdrantKnowledgeStore(
        collection_name="test collection",
    )
    nodes = [
        KnowledgeNode(
            node_id="1",
            embedding=[1, 1, 1],
            node_type="text",
            text_content="mock node",
        ),
        KnowledgeNode(
            node_id="2",
            embedding=[2, 2, 2],
            node_type="text",
            text_content="mock node",
        ),
    ]

    mock_client.upload_points.side_effect = RuntimeError(
        "mock error from qdrant"
    )

    with pytest.raises(
        LoadNodeError,
        match="Loading nodes into collection 'test collection' failed: mock error from qdrant",
    ):
        knowledge_store.load_nodes(nodes)


@patch("qdrant_client.QdrantClient")
def test_private_ensure_collection_exists(
    mock_qdrant_client_class: MagicMock,
) -> None:
    mock_client = MagicMock()
    mock_qdrant_client_class.return_value = mock_client
    knowledge_store = QdrantKnowledgeStore(
        collection_name="test collection",
    )
    mock_client.collection_exists.return_value = True

    with does_not_raise():
        knowledge_store._ensure_collection_exists()


@patch("qdrant_client.QdrantClient")
def test_private_ensure_collection_exists_raises_not_found(
    mock_qdrant_client_class: MagicMock,
) -> None:
    mock_client = MagicMock()
    mock_qdrant_client_class.return_value = mock_client
    knowledge_store = QdrantKnowledgeStore(
        collection_name="test collection",
    )
    mock_client.collection_exists.return_value = False

    with pytest.raises(
        KnowledgeStoreNotFoundError,
        match="Collection 'test collection' does not exist.",
    ):
        knowledge_store._ensure_collection_exists()


@patch("qdrant_client.QdrantClient")
def test_private_create_collection(
    mock_qdrant_client_class: MagicMock,
) -> None:
    from qdrant_client.models import Distance, VectorParams

    mock_client = MagicMock()
    mock_qdrant_client_class.return_value = mock_client
    knowledge_store = QdrantKnowledgeStore(
        collection_name="test collection",
    )
    distance = Distance(knowledge_store.collection_distance)

    # act
    knowledge_store._create_collection(
        collection_name="test collection", vector_size=100, distance=distance
    )

    mock_client.create_collection.assert_called_once_with(
        collection_name="test collection",
        vectors_config=VectorParams(size=100, distance=distance),
    )


@patch("qdrant_client.QdrantClient")
def test_private_create_collection_raises_invalid_distance_error(
    mock_qdrant_client_class: MagicMock,
) -> None:
    mock_client = MagicMock()
    mock_qdrant_client_class.return_value = mock_client
    knowledge_store = QdrantKnowledgeStore(
        collection_name="test collection",
    )

    # act
    with pytest.raises(InvalidDistanceError):
        knowledge_store._create_collection(
            collection_name="test collection",
            vector_size=100,
            distance="invalid distance",
        )


@patch("qdrant_client.QdrantClient")
def test_private_create_collection_raises_qdrant_error(
    mock_qdrant_client_class: MagicMock,
) -> None:
    mock_client = MagicMock()
    mock_qdrant_client_class.return_value = mock_client
    knowledge_store = QdrantKnowledgeStore(
        collection_name="test collection",
    )

    mock_client.create_collection.side_effect = RuntimeError(
        "mock qdrant error"
    )

    # act
    with pytest.raises(
        KnowledgeStoreError,
        match="Failed to create collection: mock qdrant error",
    ):
        knowledge_store._create_collection(
            collection_name="test collection",
            vector_size=100,
            distance=knowledge_store.collection_distance,
        )


@patch.object(QdrantKnowledgeStore, "_ensure_collection_exists")
@patch("qdrant_client.QdrantClient")
def test_retrieve(
    mock_qdrant_client_class: MagicMock,
    mock_ensure_collection_exists: MagicMock,
) -> None:
    from qdrant_client.conversions.common_types import QueryResponse
    from qdrant_client.http.models import ScoredPoint

    mock_client = MagicMock()
    mock_qdrant_client_class.return_value = mock_client
    knowledge_store = QdrantKnowledgeStore(
        collection_name="test collection",
    )

    test_node = KnowledgeNode(
        node_id="1",
        embedding=[1, 1, 1],
        node_type="text",
        text_content="mock node",
    )
    test_pt = ScoredPoint(
        id="1", score=0.42, version=1, payload=test_node.model_dump()
    )
    test_query_response = QueryResponse(points=[test_pt])
    mock_client.query_points.return_value = test_query_response

    # act
    retrieval_res = knowledge_store.retrieve(query_emb=[1, 1, 1], top_k=5)

    # assert
    expected = [
        _convert_scored_point_to_knowledge_node_and_score_tuple(test_pt)
    ]
    assert expected == retrieval_res
    mock_ensure_collection_exists.assert_called_once()


@patch.object(QdrantKnowledgeStore, "_ensure_collection_exists")
@patch("qdrant_client.QdrantClient")
def test_retrieve_raises_error(
    mock_qdrant_client_class: MagicMock,
    mock_ensure_collection_exists: MagicMock,
) -> None:
    mock_client = MagicMock()
    mock_qdrant_client_class.return_value = mock_client
    knowledge_store = QdrantKnowledgeStore(
        collection_name="test collection",
    )

    mock_client.query_points.side_effect = RuntimeError("mock qdrant error")

    # act
    with pytest.raises(
        KnowledgeStoreError,
        match="Failed to retrieve from collection 'test collection': mock qdrant error",
    ):
        knowledge_store.retrieve(query_emb=[1, 1, 1], top_k=5)

    mock_ensure_collection_exists.assert_called_once()


def test_persist_raises_error() -> None:
    knowledge_store = QdrantKnowledgeStore(
        collection_name="test collection",
    )

    with pytest.raises(
        NotImplementedError,
        match=re.escape(
            "`persist()` is not available in QdrantKnowledgeStore."
        ),
    ):
        knowledge_store.persist()


def test_load_raises_error() -> None:
    knowledge_store = QdrantKnowledgeStore(
        collection_name="test collection",
    )

    msg = (
        "`load()` is not available in QdrantKnowledgeStore. "
        "Data is automatically persisted and loaded from the Qdrant server."
    )
    with pytest.raises(NotImplementedError, match=re.escape(msg)):
        knowledge_store.load()


@patch.object(QdrantKnowledgeStore, "_ensure_collection_exists")
@patch("qdrant_client.QdrantClient")
def test_delete_node(
    mock_qdrant_client_class: MagicMock,
    mock_ensure_collection_exists: MagicMock,
) -> None:
    from qdrant_client.http.models import (
        FieldCondition,
        Filter,
        MatchValue,
        UpdateResult,
        UpdateStatus,
    )

    mock_client = MagicMock()
    mock_qdrant_client_class.return_value = mock_client
    knowledge_store = QdrantKnowledgeStore(
        collection_name="test collection",
    )

    test_update_result = UpdateResult(status=UpdateStatus.COMPLETED)

    mock_client.delete.return_value = test_update_result

    # act
    knowledge_store.delete_node(node_id="1")

    # assert
    mock_client.delete.assert_called_once_with(
        collection_name="test collection",
        points_selector=Filter(
            must=[FieldCondition(key="node_id", match=MatchValue(value="1"))]
        ),
    )
    mock_ensure_collection_exists.assert_called_once()


@patch.object(QdrantKnowledgeStore, "_ensure_collection_exists")
@patch("qdrant_client.QdrantClient")
def test_delete_node_raises_error(
    mock_qdrant_client_class: MagicMock,
    mock_ensure_collection_exists: MagicMock,
) -> None:
    mock_client = MagicMock()
    mock_qdrant_client_class.return_value = mock_client
    knowledge_store = QdrantKnowledgeStore(
        collection_name="test collection",
    )

    mock_client.delete.side_effect = RuntimeError("mock qdrant error")

    # act
    with pytest.raises(
        KnowledgeStoreError,
        match="Failed to delete node: '1' from collection 'test collection'",
    ):
        knowledge_store.delete_node(node_id="1")


@patch.object(QdrantKnowledgeStore, "_ensure_collection_exists")
@patch("qdrant_client.QdrantClient")
def test_clear(
    mock_qdrant_client_class: MagicMock,
    mock_ensure_collection_exists: MagicMock,
) -> None:
    mock_client = MagicMock()
    mock_qdrant_client_class.return_value = mock_client
    knowledge_store = QdrantKnowledgeStore(
        collection_name="test collection",
    )

    # act
    with does_not_raise():
        knowledge_store.clear()

    # assert
    mock_ensure_collection_exists.assert_called_once()
    mock_client.delete_collection.assert_called_once_with(
        collection_name="test collection"
    )


@patch.object(QdrantKnowledgeStore, "_ensure_collection_exists")
@patch("qdrant_client.QdrantClient")
def test_clear_raises_error(
    mock_qdrant_client_class: MagicMock,
    mock_ensure_collection_exists: MagicMock,
) -> None:
    mock_client = MagicMock()
    mock_qdrant_client_class.return_value = mock_client
    knowledge_store = QdrantKnowledgeStore(
        collection_name="test collection",
    )

    # act
    mock_client.delete_collection.side_effect = RuntimeError(
        "mock qdrant error"
    )

    with pytest.raises(
        KnowledgeStoreError,
        match="Failed to delete collection 'test collection': mock qdrant error",
    ):
        knowledge_store.clear()

    # assert
    mock_ensure_collection_exists.assert_called_once()
    mock_client.delete_collection.assert_called_once_with(
        collection_name="test collection"
    )


@patch.object(QdrantKnowledgeStore, "_ensure_collection_exists")
@patch("qdrant_client.QdrantClient")
def test_count(
    mock_qdrant_client_class: MagicMock,
    mock_ensure_collection_exists: MagicMock,
) -> None:
    from qdrant_client.http.models import CountResult

    mock_client = MagicMock()
    mock_qdrant_client_class.return_value = mock_client
    knowledge_store = QdrantKnowledgeStore(
        collection_name="test collection",
    )

    test_count_result = CountResult(count=10)
    mock_client.count.return_value = test_count_result

    # act
    res = knowledge_store.count

    # assert
    assert res == 10
    mock_ensure_collection_exists.assert_called_once()
    mock_client.count.assert_called_once_with(
        collection_name="test collection"
    )


@patch.object(QdrantKnowledgeStore, "_ensure_collection_exists")
@patch("qdrant_client.QdrantClient")
def test_count_raises_error(
    mock_qdrant_client_class: MagicMock,
    mock_ensure_collection_exists: MagicMock,
) -> None:
    mock_client = MagicMock()
    mock_qdrant_client_class.return_value = mock_client
    knowledge_store = QdrantKnowledgeStore(
        collection_name="test collection",
    )

    mock_client.count.side_effect = RuntimeError("mock qdrant error")

    # act
    with pytest.raises(
        KnowledgeStoreError,
        match="Failed to get vector count for collection 'test collection': mock qdrant error",
    ):
        knowledge_store.count

    # assert
    mock_ensure_collection_exists.assert_called_once()
    mock_client.count.assert_called_once_with(
        collection_name="test collection"
    )


@patch.object(QdrantKnowledgeStore, "_collection_exists")
@patch.object(QdrantKnowledgeStore, "_create_collection")
@patch("qdrant_client.QdrantClient")
def test_private_check_if_collection_exists_otherwise_create_one(
    mock_qdrant_client_class: MagicMock,
    mock_create_collection: MagicMock,
    mock_collection_exists: MagicMock,
) -> None:
    mock_client = MagicMock()
    mock_qdrant_client_class.return_value = mock_client
    knowledge_store = QdrantKnowledgeStore(
        collection_name="test collection",
    )
    mock_collection_exists.return_value = False

    # act
    knowledge_store._check_if_collection_exists_otherwise_create_one(
        vector_size=10
    )

    mock_collection_exists.assert_called_once()
    mock_create_collection.assert_called_once_with(
        collection_name="test collection", vector_size=10, distance="Cosine"
    )


@patch.object(QdrantKnowledgeStore, "_collection_exists")
@patch.object(QdrantKnowledgeStore, "_create_collection")
@patch("qdrant_client.QdrantClient")
def test_private_check_if_collection_exists_otherwise_create_one_raises_error(
    mock_qdrant_client_class: MagicMock,
    mock_create_collection: MagicMock,
    mock_collection_exists: MagicMock,
) -> None:
    mock_client = MagicMock()
    mock_qdrant_client_class.return_value = mock_client
    knowledge_store = QdrantKnowledgeStore(
        collection_name="test collection",
    )
    mock_collection_exists.return_value = False
    mock_create_collection.side_effect = KnowledgeStoreError("mock error")

    # act
    with pytest.raises(
        KnowledgeStoreError,
        match="Failed to create new collection: 'test collection'",
    ):
        knowledge_store._check_if_collection_exists_otherwise_create_one(
            vector_size=10
        )

    mock_collection_exists.assert_called_once()
    mock_create_collection.assert_called_once_with(
        collection_name="test collection", vector_size=10, distance="Cosine"
    )



"""Base FL Task"""

from abc import ABC, abstractmethod
from typing import Any, Callable

from flwr.client.client import Client
from flwr.server.server import Server
from pydantic import BaseModel, ConfigDict
from typing_extensions import Self

from fed_rag.exceptions import MissingFLTaskConfig


class BaseFLTaskConfig(BaseModel):
    pass


class BaseFLTask(BaseModel, ABC):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    @property
    @abstractmethod
    def training_loop(self) -> Callable:
        ...

    @classmethod
    @abstractmethod
    def from_configs(
        cls, trainer_cfg: BaseFLTaskConfig, tester_cfg: Any
    ) -> Self:
        ...

    @classmethod
    @abstractmethod
    def from_trainer_and_tester(
        cls, trainer: Callable, tester: Callable
    ) -> Self:
        try:
            trainer_cfg = getattr(trainer, "__fl_task_trainer_config")
        except AttributeError:
            msg = (
                "`__fl_task_trainer_config` has not been set on training loop. Make "
                "sure to decorate your training loop with the appropriate "
                "decorator."
            )
            raise MissingFLTaskConfig(msg)

        try:
            tester_cfg = getattr(tester, "__fl_task_tester_config")
        except AttributeError:
            msg = (
                "`__fl_task_tester_config` has not been set on tester callable. Make "
                "sure to decorate your tester with the appropriate decorator."
            )
            raise MissingFLTaskConfig(msg)
        return cls.from_configs(trainer_cfg, tester_cfg)

    @abstractmethod
    def simulate(self, num_clients: int, **kwargs: Any) -> Any:
        """Simulate the FL task.

        Either use flwr's simulation tools, or create our own here.
        """
        ...

    @abstractmethod
    def server(self, **kwargs: Any) -> Server:
        """Create a flwr.Server object."""
        ...

    @abstractmethod
    def client(self, **kwargs: Any) -> Client:
        """Create a flwr.Client object."""
        ...



import pytest
import torch


@pytest.fixture()
def input_and_target_ids() -> tuple[list[torch.Tensor], list[torch.Tensor]]:
    input_ids = [torch.zeros(3)] * 3
    target_ids = [torch.ones(3)] * 3
    return input_ids, target_ids



from fed_rag.types.rag_system import RAGSystem

from .conftest import MockRetrieverTrainer, MockTrainer


def test_init(mock_rag_system: RAGSystem) -> None:
    trainer = MockTrainer(
        rag_system=mock_rag_system,
        train_dataset=[{"query": "mock example", "response": "mock response"}],
    )

    assert trainer.rag_system == mock_rag_system


def test_retriever_trainer_with_dual_encoder_retriever(
    mock_rag_system_dual_encoder: RAGSystem,
) -> None:
    trainer = MockRetrieverTrainer(
        rag_system=mock_rag_system_dual_encoder,
        train_dataset=[{"query": "mock example", "response": "mock response"}],
    )

    assert trainer.rag_system == mock_rag_system_dual_encoder
    assert (
        trainer.model == mock_rag_system_dual_encoder.retriever.query_encoder
    )



"""Base Knowledge Store"""

from abc import ABC, abstractmethod

from pydantic import BaseModel, ConfigDict, Field

from fed_rag.types.knowledge_node import KnowledgeNode

DEFAULT_KNOWLEDGE_STORE_NAME = "default"


class BaseKnowledgeStore(BaseModel, ABC):
    """Base Knowledge Store Class."""

    model_config = ConfigDict(arbitrary_types_allowed=True)
    name: str = Field(
        description="Name of Knowledge Store used for caching and loading.",
        default=DEFAULT_KNOWLEDGE_STORE_NAME,
    )

    @abstractmethod
    def load_node(self, node: KnowledgeNode) -> None:
        """Load a KnowledgeNode into the KnowledgeStore."""

    @abstractmethod
    def load_nodes(self, nodes: list[KnowledgeNode]) -> None:
        """Load multiple KnowledgeNodes in batch."""

    @abstractmethod
    def retrieve(
        self, query_emb: list[float], top_k: int
    ) -> list[tuple[float, KnowledgeNode]]:
        """Retrieve top-k nodes from KnowledgeStore against a provided user query.

        Returns:
            A list of tuples where the first element represents the similarity score
            of the node to the query, and the second element is the node itself.
        """

    @abstractmethod
    def delete_node(self, node_id: str) -> bool:
        """Remove a node from the KnowledgeStore by ID, returning success status."""

    @abstractmethod
    def clear(self) -> None:
        """Clear all nodes from the KnowledgeStore."""

    @property
    @abstractmethod
    def count(self) -> int:
        """Return the number of nodes in the store."""

    @abstractmethod
    def persist(self) -> None:
        """Save the KnowledgeStore nodes to a permanent storage."""

    @abstractmethod
    def load(self) -> None:
        """
        Load the KnowledgeStore nodes from a permanent storage using `name`.

        Args:
            ks_id: The id of the knowledge store to load.
        """



from fed_rag.types.knowledge_node import KnowledgeNode
from fed_rag.types.rag_system import SourceNode


def test_getattr_sourcenode_wraps_knowledge_node() -> None:
    # arrange
    knowledge_node = KnowledgeNode(
        embedding=[0.1, 0.2],
        node_type="text",
        text_content="fake text context",
        metadata={"some_field": 12},
    )

    # act
    source_node = SourceNode(score=0.42, node=knowledge_node)

    # assert
    assert source_node.score == 0.42
    assert source_node.text_content == knowledge_node.text_content
    assert source_node.node_type == knowledge_node.node_type
    assert source_node.node_id == knowledge_node.node_id
    assert source_node.metadata == knowledge_node.metadata



"""Decorators"""

from .tester import TesterDecorators
from .trainer import TrainerDecorators


class Federate:
    def __init__(self, trainer: TrainerDecorators, tester: TesterDecorators):
        self.trainer = trainer
        self.tester = tester


federate = Federate(trainer=TrainerDecorators(), tester=TesterDecorators())


__all__ = ["federate"]



"""HuggingFace Data Collator For Retrieval-Augmented Generator Training"""

from typing import TYPE_CHECKING, Any

import torch
from pydantic import Field

from fed_rag.base.data_collator import BaseDataCollator
from fed_rag.exceptions import DataCollatorError, MissingExtraError
from fed_rag.types.rag_system import RAGSystem
from fed_rag.utils.huggingface import _validate_rag_system

try:
    from transformers.data.data_collator import DataCollatorMixin

    _has_huggingface = True
except ModuleNotFoundError:
    _has_huggingface = False

    # Create a dummy class with a different name to avoid the redefinition
    class _DummyDataCollatorMixin:
        """Dummy placeholder when transformers is not available."""

        pass

    DataCollatorMixin = _DummyDataCollatorMixin  # type: ignore

if TYPE_CHECKING:  # pragma: no cover
    from transformers import PreTrainedTokenizer


DEFAULT_EXAMPLE_TEMPLATE = """
You are a helpful assistant. Given the user's question, provide a succinct
and accurate response. If context is provided, use it in your answer if it helps
you to create the most accurate response.

<warning>
Only use the the provided context if its relevant to answer the question. Otherwise,
ignore it and use your parametric knowledge to answer the question.
</warning>

<question>
{query}
</question>

<context>
{context}
</context>

<response>
{response}
</response>
"""


class DataCollatorForRALT(DataCollatorMixin, BaseDataCollator):
    """A HuggingFace DataCollator for LM-Supervised Retrieval."""

    example_template: str = Field(default=DEFAULT_EXAMPLE_TEMPLATE)
    default_return_tensors: str = Field(default="pt")

    def __init__(
        self,
        rag_system: RAGSystem,
        example_template: str | None = None,
        default_return_tensors: str = "pt",
        **kwargs: Any,
    ):
        if not _has_huggingface:
            msg = (
                f"`{self.__class__.__name__}` requires `huggingface` extra to be installed. "
                "To fix please run `pip install fed-rag[huggingface]`."
            )
            raise MissingExtraError(msg)

        _validate_rag_system(rag_system)

        example_template = example_template or DEFAULT_EXAMPLE_TEMPLATE

        super().__init__(
            rag_system=rag_system,
            example_template=example_template,
            default_return_tensors=default_return_tensors,
            **kwargs,
        )

    def _apply_padding(
        self,
        max_length: int,
        inputs_list: list[list[int]],
        attention_mask_list: list[list[int]],
        tokenizer: "PreTrainedTokenizer",
    ) -> dict[str, torch.Tensor]:
        """Applys left padding for causal lm modelling."""

        # First convert lists to tensors if not already
        input_ids_tensors = [torch.tensor(ids) for ids in inputs_list]
        attention_mask_tensors = [
            torch.tensor(mask) for mask in attention_mask_list
        ]
        labels_tensors = [
            torch.tensor(ids) for ids in inputs_list
        ]  # Labels are the same as input_ids for causal LM

        # Get pad token ID
        if tokenizer.pad_token is not None:
            if tokenizer.pad_token_id < 0:
                raise DataCollatorError(
                    "Asking to pad but the tokenizer has a value for pad_token_id < 0."
                )
            pad_token_id = tokenizer.pad_token_id
        else:
            if tokenizer.eos_token_id is not None:
                pad_token_id = tokenizer.eos_token_id
            else:
                raise DataCollatorError(
                    "Asking to pad but the tokenizer does not have a padding token "
                    "nor an eos token that can potentially be used in its place."
                )

        # Create padded tensors
        padded_input_ids = []
        padded_attention_mask = []
        padded_labels = []

        for input_ids, attention_mask, labels in zip(
            input_ids_tensors, attention_mask_tensors, labels_tensors
        ):
            # Calculate padding needed
            pad_len = max_length - len(input_ids)

            if pad_len > 0:
                # Create padding tensors
                padding = torch.full(
                    (pad_len,), pad_token_id, dtype=input_ids.dtype
                )
                mask_padding = torch.zeros(pad_len, dtype=attention_mask.dtype)
                label_padding = torch.full(
                    (pad_len,), -100, dtype=labels.dtype
                )  # -100 to ignore in loss calculation

                # Apply left padding
                padded_input = torch.cat([padding, input_ids])
                padded_mask = torch.cat([mask_padding, attention_mask])
                padded_label = torch.cat([label_padding, labels])
            else:
                # No padding needed
                padded_input = input_ids
                padded_mask = attention_mask
                padded_label = labels

            padded_input_ids.append(padded_input)
            padded_attention_mask.append(padded_mask)
            padded_labels.append(padded_label)

        # Stack into batch tensors
        return {
            "input_ids": torch.stack(padded_input_ids),
            "attention_mask": torch.stack(padded_attention_mask),
            "labels": torch.stack(padded_labels),
        }

    def __call__(
        self, features: list[dict[str, Any]], return_tensors: str | None = None
    ) -> dict[str, Any]:
        """Use the features of the dataset in order to get the `input_ids` and `labels`.

        Steps:
            1. process the features using the RAG system and example template to create
               the retrieval-augmented lm fine-tuning text
            2. apply padding and get required ~torch.Tensors


        Args:
            features (list[Any]): Should contain a 'query' and 'response' field.
            return_tensors (_type_, optional): supports right now only 'pt'

        Returns:
            dict[str, Any]: a dictionary of ~torch.Tensors with keys 'input_ids'
                and 'labels'

        Note that each ('query', 'response') pair generates rag_system.config.top_k
        fine-tuning instance for RALT.
        """
        return_tensors = (
            return_tensors if return_tensors else self.default_return_tensors
        )
        if return_tensors != "pt":
            raise DataCollatorError(
                f"Framework '{return_tensors}' not recognized!"
            )

        # STEP 1 — use rag system to build the RALT fine-tuning texts
        finetuning_instances = []
        inputs_list = []
        attention_mask_list = []
        max_length = 0
        for example in features:
            # retrieve
            source_nodes = self.rag_system.retrieve(query=example["query"])
            total_sum_scores = sum(s.score for s in source_nodes)

            # parallel in-context retrieval-augmentation creates
            # top_k separated finetuning instances
            for source in source_nodes:
                finetune_instance_text = self.example_template.format(
                    query=example["query"],
                    response=example["response"],
                    context=source.node.get_content()["text_content"],
                )
                finetuning_instances.append(finetune_instance_text)
                _weight = source.score / total_sum_scores

                # tokenize to get input_ids and target_ids
                tokenizer = self.rag_system.generator.tokenizer

                encode_result = tokenizer.encode(finetune_instance_text)
                input_ids = encode_result["input_ids"]
                attention_mask = encode_result["attention_mask"]

                current_input_len = len(input_ids)
                if current_input_len > max_length:
                    max_length = current_input_len

                inputs_list.append(input_ids)
                attention_mask_list.append(attention_mask)

        # padding — apply left padding
        padded_features = self._apply_padding(
            max_length=max_length,
            inputs_list=inputs_list,
            attention_mask_list=attention_mask_list,
            tokenizer=tokenizer.unwrapped,
        )

        return padded_features



"""PyTorch Trainer Inspector"""

import inspect
from typing import Any, Callable

from fed_rag.exceptions import (
    InvalidReturnType,
    MissingDataParam,
    MissingMultipleDataParams,
    MissingNetParam,
)
from fed_rag.inspectors.common import TrainerSignatureSpec
from fed_rag.types import TrainResult


def inspect_trainer_signature(fn: Callable) -> TrainerSignatureSpec:
    sig = inspect.signature(fn)

    # validate return type
    return_type = sig.return_annotation
    if (return_type is Any) or not issubclass(return_type, TrainResult):
        msg = "Trainer should return a fed_rag.types.TrainResult or a subclsas of it."
        raise InvalidReturnType(msg)

    # inspect fn params
    extra_train_kwargs = []
    net_param = None
    train_data_param = None
    val_data_param = None
    net_parameter_class_name = None

    for name, t in sig.parameters.items():
        if name in ("self", "cls"):
            continue

        if type_name := getattr(t.annotation, "__name__", None):
            if type_name == "Module" and net_param is None:
                net_param = name
                net_parameter_class_name = type_name
                continue

            if type_name == "DataLoader" and train_data_param is None:
                train_data_param = name
                continue

            if type_name == "DataLoader" and val_data_param is None:
                val_data_param = name
                continue

        extra_train_kwargs.append(name)

    if net_param is None:
        msg = (
            "Inspection failed to find a model param. "
            "For PyTorch this param must have type `nn.Module`."
        )
        raise MissingNetParam(msg)

    if train_data_param is None:
        msg = (
            "Inspection failed to find two data params for train and val datasets."
            "For PyTorch these params must be of type `torch.utils.data.DataLoader`"
        )
        raise MissingMultipleDataParams(msg)

    if val_data_param is None:
        msg = (
            "Inspection found one data param but failed to find another. "
            "Two data params are required for train and val datasets."
            "For PyTorch these params must be of type `torch.utils.data.DataLoader`"
        )
        raise MissingDataParam(msg)

    spec = TrainerSignatureSpec(
        net_parameter=net_param,
        train_data_param=train_data_param,
        val_data_param=val_data_param,
        extra_train_kwargs=extra_train_kwargs,
        net_parameter_class_name=net_parameter_class_name,
    )
    return spec



"""In Memory Knowledge Store"""

from pathlib import Path
from typing import Any, Dict, cast

import numpy as np
import pyarrow as pa
import pyarrow.parquet as pq
from pydantic import Field, PrivateAttr, model_serializer
from typing_extensions import Self

from fed_rag.base.knowledge_store import BaseKnowledgeStore
from fed_rag.exceptions.knowledge_stores import KnowledgeStoreNotFoundError
from fed_rag.knowledge_stores.mixins import ManagedMixin
from fed_rag.types.knowledge_node import KnowledgeNode

DEFAULT_CACHE_DIR = ".fed_rag/data_cache/"
DEFAULT_TOP_K = 2


def _get_top_k_nodes(
    nodes: list[KnowledgeNode],
    query_emb: list[float],
    top_k: int = DEFAULT_TOP_K,
) -> list[tuple[str, float]]:
    """Retrieves the top-k similar nodes against query.

    Returns:
        list[tuple[float, str]] — the node_ids and similarity scores of top-k nodes
    """

    def cosine_sim(a: list[float], b: list[float]) -> float:
        """Compute cosine similarity between two embeddings."""
        np_a = np.array(a)
        np_b = np.array(b)
        cosine_sim: float = np.dot(np_a, np_b) / (
            np.linalg.norm(np_a) * np.linalg.norm(np_b)
        )
        return cosine_sim

    scores = [
        (node.node_id, cosine_sim(node.embedding, query_emb)) for node in nodes
    ]
    scores.sort(key=lambda tup: tup[1], reverse=True)
    return scores[:top_k]


class InMemoryKnowledgeStore(BaseKnowledgeStore):
    """InMemoryKnowledgeStore Class."""

    cache_dir: str = Field(default=DEFAULT_CACHE_DIR)
    _data: dict[str, KnowledgeNode] = PrivateAttr(default_factory=dict)

    @classmethod
    def from_nodes(cls, nodes: list[KnowledgeNode], **kwargs: Any) -> Self:
        instance = cls(**kwargs)
        instance.load_nodes(nodes)
        return instance

    def load_node(self, node: KnowledgeNode) -> None:
        if node.node_id not in self._data:
            self._data[node.node_id] = node

    def load_nodes(self, nodes: list[KnowledgeNode]) -> None:
        for node in nodes:
            self.load_node(node)

    def retrieve(
        self, query_emb: list[float], top_k: int
    ) -> list[tuple[float, KnowledgeNode]]:
        all_nodes = list(self._data.values())
        node_ids_and_scores = _get_top_k_nodes(
            nodes=all_nodes, query_emb=query_emb, top_k=top_k
        )
        return [(el[1], self._data[el[0]]) for el in node_ids_and_scores]

    def delete_node(self, node_id: str) -> bool:
        if node_id in self._data:
            del self._data[node_id]
            return True
        else:
            return False

    def clear(self) -> None:
        self._data = {}

    @property
    def count(self) -> int:
        return len(self._data)

    @model_serializer(mode="wrap")
    def custom_model_dump(self, handler: Any) -> Dict[str, Any]:
        data = handler(self)
        data = cast(Dict[str, Any], data)
        # include _data in serialization
        if self._data:
            data["_data"] = self._data
        return data  # type: ignore[no-any-return]

    def persist(self) -> None:
        serialized_model = self.model_dump()
        data_values = list(serialized_model["_data"].values())

        parquet_table = pa.Table.from_pylist(data_values)

        filename = Path(self.cache_dir) / f"{self.name}.parquet"
        Path(filename).parent.mkdir(parents=True, exist_ok=True)
        pq.write_table(parquet_table, filename)

    def load(self) -> None:
        filename = Path(self.cache_dir) / f"{self.name}.parquet"
        if not filename.exists():
            msg = f"Knowledge store '{self.name}' not found at expected location: {filename}"
            raise KnowledgeStoreNotFoundError(msg)

        parquet_data = pq.read_table(filename).to_pylist()
        nodes = [KnowledgeNode(**data) for data in parquet_data]
        self.load_nodes(nodes)


class ManagedInMemoryKnowledgeStore(ManagedMixin, InMemoryKnowledgeStore):
    def persist(self) -> None:
        serialized_model = self.model_dump()
        data_values = list(serialized_model["_data"].values())

        parquet_table = pa.Table.from_pylist(data_values)

        filename = Path(self.cache_dir) / self.name / f"{self.ks_id}.parquet"
        Path(filename).parent.mkdir(parents=True, exist_ok=True)
        pq.write_table(parquet_table, filename)

    @classmethod
    def from_name_and_id(
        cls, name: str, ks_id: str, cache_dir: str | None = None
    ) -> Self:
        cache_dir = cache_dir if cache_dir else DEFAULT_CACHE_DIR
        filename = Path(cache_dir) / name / f"{ks_id}.parquet"
        if not filename.exists():
            msg = f"Knowledge store '{name}/{ks_id}' not found at expected location: {filename}"
            raise KnowledgeStoreNotFoundError(msg)

        parquet_data = pq.read_table(filename).to_pylist()
        nodes = [KnowledgeNode(**data) for data in parquet_data]
        knowledge_store = ManagedInMemoryKnowledgeStore.from_nodes(
            nodes, name=name, cache_dir=cache_dir
        )
        # set id
        knowledge_store.ks_id = ks_id
        return knowledge_store



"""Exceptions for FL Tasks."""

from .core import FedRAGError


class FLTaskError(FedRAGError):
    """Base fl task error for all fl-task-related exceptions."""

    pass


class MissingFLTaskConfig(FLTaskError):
    pass


class MissingRequiredNetParam(FLTaskError):
    """Raised when invoking fl_task.server without passing the specified model/net param."""

    pass


class NetTypeMismatch(FLTaskError):
    """Raised when a `trainer` and `tester` spec have differing `net_parameter_class_name`.

    This indicates that the these methods have different types for the `net_parameter`.
    """

    pass



"""HuggingFace RAG Finetuning Dataset"""

from typing_extensions import Self

from fed_rag.exceptions.common import MissingExtraError

# check if huggingface extra was installed
try:
    from datasets import Dataset
except ModuleNotFoundError:
    msg = (
        "`HuggingFaceRAGFinetuningDataset` requires the `huggingface` extra to be installed. "
        "To fix please run `pip install fed-rag[huggingface]`."
    )
    raise MissingExtraError(msg)


class HuggingFaceRAGFinetuningDataset(Dataset):
    """Thin wrapper over ~datasets.Dataset."""

    @classmethod
    def from_inputs(
        cls,
        input_ids: list[list[int]],
        target_ids: list[list[int]],
        attention_mask: list[list[int]],
    ) -> Self:
        return cls.from_dict(  # type: ignore[no-any-return]
            {
                "input_ids": input_ids,
                "target_ids": target_ids,
                "attention_mask": attention_mask,
            }
        )



from .tester import TesterSignatureSpec, inspect_tester_signature
from .trainer import TrainerSignatureSpec, inspect_trainer_signature

__all__ = [
    "TesterSignatureSpec",
    "TrainerSignatureSpec",
    "inspect_tester_signature",
    "inspect_trainer_signature",
]



"""Base Retriever"""

from abc import ABC, abstractmethod
from typing import Any

import torch
from pydantic import BaseModel, ConfigDict


class BaseRetriever(BaseModel, ABC):
    """Base Retriever Class."""

    model_config = ConfigDict(arbitrary_types_allowed=True)

    @abstractmethod
    def encode_query(
        self, query: str | list[str], **kwargs: Any
    ) -> torch.Tensor:
        """Encode query."""

    @abstractmethod
    def encode_context(
        self, context: str | list[str], **kwargs: Any
    ) -> torch.Tensor:
        """Encode context."""

    @property
    @abstractmethod
    def encoder(self) -> torch.nn.Module | None:
        """PyTorch model associated with the encoder associated with retriever."""

    @property
    @abstractmethod
    def query_encoder(self) -> torch.nn.Module | None:
        """PyTorch model associated with the query encoder associated with retriever."""

    @property
    @abstractmethod
    def context_encoder(self) -> torch.nn.Module | None:
        """PyTorch model associated with the context encoder associated with retriever."""



from unittest.mock import MagicMock, patch

import torch

from fed_rag.base.generator import BaseGenerator
from fed_rag.base.retriever import BaseRetriever
from fed_rag.knowledge_stores.in_memory import InMemoryKnowledgeStore
from fed_rag.types.knowledge_node import KnowledgeNode
from fed_rag.types.rag_system import RAGConfig, RAGSystem, SourceNode

from .conftest import MockGenerator, MockRetriever


def test_rag_system_init(
    mock_generator: BaseGenerator,
    mock_retriever: BaseRetriever,
    knowledge_nodes: list[KnowledgeNode],
) -> None:
    knowledge_store = InMemoryKnowledgeStore.from_nodes(nodes=knowledge_nodes)
    rag_config = RAGConfig(
        top_k=2,
    )
    rag_system = RAGSystem(
        generator=mock_generator,
        retriever=mock_retriever,
        knowledge_store=knowledge_store,
        rag_config=rag_config,
    )
    assert rag_system.knowledge_store == knowledge_store
    assert rag_system.rag_config == rag_config
    assert rag_system.generator == mock_generator
    assert rag_system.retriever == mock_retriever


def test_rag_system_with_dual_encoder_init(
    mock_generator: BaseGenerator,
    mock_dual_retriever: BaseRetriever,
    knowledge_nodes: list[KnowledgeNode],
) -> None:
    knowledge_store = InMemoryKnowledgeStore.from_nodes(nodes=knowledge_nodes)
    rag_config = RAGConfig(
        top_k=2,
    )
    rag_system = RAGSystem(
        generator=mock_generator,
        retriever=mock_dual_retriever,
        knowledge_store=knowledge_store,
        rag_config=rag_config,
    )
    assert rag_system.knowledge_store == knowledge_store
    assert rag_system.rag_config == rag_config
    assert rag_system.generator == mock_generator
    assert rag_system.retriever == mock_dual_retriever


@patch.object(RAGSystem, "generate")
@patch.object(RAGSystem, "_format_context")
@patch.object(RAGSystem, "retrieve")
def test_rag_system_query(
    mock_retrieve: MagicMock,
    mock_format_context: MagicMock,
    mock_generate: MagicMock,
    mock_generator: BaseGenerator,
    mock_retriever: BaseRetriever,
    knowledge_nodes: list[KnowledgeNode],
) -> None:
    # arrange mocks
    source_nodes = [
        SourceNode(score=0.99, node=knowledge_nodes[0]),
        SourceNode(score=0.85, node=knowledge_nodes[1]),
    ]
    mock_retrieve.return_value = source_nodes
    mock_format_context.return_value = "fake context"
    mock_generate.return_value = "fake generation response"

    # build rag system
    knowledge_store = InMemoryKnowledgeStore.from_nodes(nodes=knowledge_nodes)
    rag_config = RAGConfig(
        top_k=2,
    )
    rag_system = RAGSystem(
        generator=mock_generator,
        retriever=mock_retriever,
        knowledge_store=knowledge_store,
        rag_config=rag_config,
    )

    # act
    rag_response = rag_system.query(query="fake query")

    # assert
    mock_retrieve.assert_called_with("fake query")
    mock_format_context.assert_called_with(source_nodes)
    mock_generate.assert_called_with(
        query="fake query", context="fake context"
    )
    assert rag_response.source_nodes == source_nodes
    assert rag_response.response == "fake generation response"
    assert str(rag_response) == "fake generation response"


@patch.object(MockRetriever, "encode_query")
def test_rag_system_retrieve(
    mock_encode_query: MagicMock,
    mock_generator: BaseGenerator,
    mock_retriever: MockRetriever,
    knowledge_nodes: list[KnowledgeNode],
) -> None:
    # arrange mocks
    mock_encode_query.return_value = torch.Tensor([1.0, 1.0, 1.0])

    # build rag system
    knowledge_store = InMemoryKnowledgeStore.from_nodes(nodes=knowledge_nodes)
    rag_config = RAGConfig(
        top_k=2,
    )
    rag_system = RAGSystem(
        generator=mock_generator,
        retriever=mock_retriever,
        knowledge_store=knowledge_store,
        rag_config=rag_config,
    )

    # expected retrieved source nodes
    expected = [
        SourceNode(score=0.8164965809277259, node=knowledge_nodes[0]),
        SourceNode(score=0.8164965809277259, node=knowledge_nodes[2]),
    ]

    # act
    result = rag_system.retrieve("fake query")

    # assert
    mock_encode_query.assert_called_with("fake query")
    assert all(res.node == exp.node for res, exp in zip(result, expected))
    assert all(
        abs(res.score - exp.score) < 1e-5 for res, exp in zip(result, expected)
    )


@patch.object(MockGenerator, "generate")
def test_rag_system_generate(
    mock_generate: MagicMock,
    mock_generator: MockGenerator,
    mock_retriever: MockRetriever,
    knowledge_nodes: list[KnowledgeNode],
) -> None:
    # arrange mocks
    mock_generate.return_value = "fake generate response"

    # build rag system
    knowledge_store = InMemoryKnowledgeStore.from_nodes(nodes=knowledge_nodes)
    rag_config = RAGConfig(
        top_k=2,
    )
    rag_system = RAGSystem(
        generator=mock_generator,
        retriever=mock_retriever,
        knowledge_store=knowledge_store,
        rag_config=rag_config,
    )

    # act
    res = rag_system.generate(query="fake query", context="fake context")

    # assert
    mock_generate.assert_called_once_with(
        query="fake query", context="fake context"
    )
    assert res == "fake generate response"


def test_rag_system_format_context(
    mock_generator: MockGenerator,
    mock_retriever: MockRetriever,
    knowledge_nodes: list[KnowledgeNode],
) -> None:
    # build rag system
    knowledge_store = InMemoryKnowledgeStore.from_nodes(nodes=knowledge_nodes)
    rag_config = RAGConfig(
        top_k=2,
    )
    rag_system = RAGSystem(
        generator=mock_generator,
        retriever=mock_retriever,
        knowledge_store=knowledge_store,
        rag_config=rag_config,
    )

    # act
    formatted_context = rag_system._format_context(
        source_nodes=knowledge_nodes
    )

    # assert
    assert formatted_context == "node 1\nnode 2\nnode 3"



"""Base Tokenizer"""

from abc import ABC, abstractmethod
from typing import Any, TypedDict

from pydantic import BaseModel, ConfigDict


class EncodeResult(TypedDict):
    input_ids: list[int]
    attention_mask: list[int] | None


class BaseTokenizer(BaseModel, ABC):
    """Base Tokenizer Class."""

    model_config = ConfigDict(arbitrary_types_allowed=True)

    @abstractmethod
    def encode(self, input: str, **kwargs: dict) -> EncodeResult:
        """Encode the input string into list of integers."""

    @abstractmethod
    def decode(self, input_ids: str, **kwargs: dict) -> str:
        """Decode the input token ids into a string."""

    @property
    @abstractmethod
    def unwrapped(self) -> Any:
        """Return the underlying tokenizer if there is one."""



from unittest.mock import MagicMock, patch

import pytest

from fed_rag.types.knowledge_node import KnowledgeNode, NodeContent


@patch("fed_rag.types.knowledge_node.uuid")
def test_text_knowledge_node_init(mock_uuid: MagicMock) -> None:
    mock_uuid.uuid4.return_value = "mock_id"
    node = KnowledgeNode(
        embedding=[0.1, 0.2, 0.3], node_type="text", text_content="mock_text"
    )

    assert node.node_id == "mock_id"
    assert node.embedding == [0.1, 0.2, 0.3]
    assert node.node_type == "text"
    assert node.text_content == "mock_text"


def test_text_knowledge_node_init_raises_validation_error() -> None:
    with pytest.raises(
        ValueError, match="NodeType == 'text', but text_content is None."
    ):
        KnowledgeNode(
            node_id="mock_id", embedding=[0.1, 0.2, 0.3], node_type="text"
        )


@patch("fed_rag.types.knowledge_node.uuid")
def test_image_knowledge_node_init(mock_uuid: MagicMock) -> None:
    mock_uuid.uuid4.return_value = "mock_id"
    node = KnowledgeNode(
        embedding=[0.1, 0.2, 0.3],
        node_type="image",
        image_content=b"mock_base64_str",
    )

    assert node.node_id == "mock_id"
    assert node.embedding == [0.1, 0.2, 0.3]
    assert node.node_type == "image"
    assert isinstance(node.image_content, bytes)
    assert node.image_content == b"mock_base64_str"


def test_image_knowledge_node_init_raises_validation_error() -> None:
    with pytest.raises(
        ValueError, match="NodeType == 'image', but image_content is None."
    ):
        KnowledgeNode(
            node_id="mock_id", embedding=[0.1, 0.2, 0.3], node_type="image"
        )


@patch("fed_rag.types.knowledge_node.uuid")
def test_multimodal_knowledge_node_init(mock_uuid: MagicMock) -> None:
    mock_uuid.uuid4.return_value = "mock_id"
    node = KnowledgeNode(
        embedding=[0.1, 0.2, 0.3],
        node_type="multimodal",
        text_content="fake content",
        image_content=b"mock_base64_str",
    )

    assert node.node_id == "mock_id"
    assert node.embedding == [0.1, 0.2, 0.3]
    assert node.text_content == "fake content"
    assert node.node_type == "multimodal"
    assert isinstance(node.image_content, bytes)
    assert node.image_content == b"mock_base64_str"


def test_multimodal_knowledge_node_init_raises_validation_error_missing_text() -> (
    None
):
    with pytest.raises(
        ValueError, match="NodeType == 'multimodal', but text_content is None."
    ):
        KnowledgeNode(
            node_id="mock_id",
            embedding=[0.1, 0.2, 0.3],
            node_type="multimodal",
        )


def test_multimodal_knowledge_node_init_raises_validation_error_missing_image() -> (
    None
):
    with pytest.raises(
        ValueError,
        match="NodeType == 'multimodal', but image_content is None.",
    ):
        KnowledgeNode(
            node_id="mock_id",
            embedding=[0.1, 0.2, 0.3],
            node_type="multimodal",
            text_content="fake content",
        )


@pytest.mark.parametrize(
    ("node", "expected_content"),
    [
        (
            KnowledgeNode(
                node_type="text",
                embedding=[0.1, 0.2],
                text_content="fake content",
            ),
            {"text_content": "fake content", "image_content": None},
        ),
        (
            KnowledgeNode(
                node_type="image",
                embedding=[0.1, 0.2],
                image_content=b"fake-base64-str",
            ),
            {"text_content": None, "image_content": b"fake-base64-str"},
        ),
        (
            KnowledgeNode(
                node_type="multimodal",
                embedding=[0.1, 0.2],
                text_content="fake content",
                image_content=b"fake-base64-str",
            ),
            {
                "text_content": "fake content",
                "image_content": b"fake-base64-str",
            },
        ),
    ],
)
def test_get_content(
    node: KnowledgeNode, expected_content: NodeContent
) -> None:
    assert node.get_content() == expected_content


def test_serialize_metadata() -> None:
    # test when metadata is None
    node = KnowledgeNode(
        node_type="text",
        embedding=[0.1, 0.2],
        text_content="content",
    )
    serialized_content = node.model_dump()
    assert not serialized_content["metadata"]

    # test when metadata is not None
    node = KnowledgeNode(
        node_type="text",
        embedding=[0.1, 0.2],
        text_content="content",
        metadata={"key1": "value1", "key2": "value2"},
    )
    serialized_content = node.model_dump()
    assert "node_id" in serialized_content
    assert (
        serialized_content["metadata"]
        == '{"key1": "value1", "key2": "value2"}'
    )


def test_deserialize_metadata() -> None:
    # test when metadata is None
    node = KnowledgeNode(
        node_type="text",
        embedding=[0.1, 0.2],
        text_content="content",
    )
    assert node.metadata == {}

    # test when metadata is a dictionary
    node = KnowledgeNode(
        node_type="text",
        embedding=[0.1, 0.2],
        text_content="content",
        metadata={"key1": "value1", "key2": "value2"},
    )
    assert node.metadata == {"key1": "value1", "key2": "value2"}

    # test when metadata is a json string
    node = KnowledgeNode(
        node_type="text",
        embedding=[0.1, 0.2],
        text_content="content",
        metadata='{"key1": "value1", "key2": "value2"}',
    )
    assert node.metadata == {"key1": "value1", "key2": "value2"}


def test_deserialize_metadata_when_empty() -> None:
    node = KnowledgeNode(
        node_type="text",
        embedding=[0.1, 0.2],
        text_content="content",
    )
    serialized = node.model_dump()

    deserialized_node = KnowledgeNode.model_validate(serialized)

    assert node == deserialized_node
    assert serialized["metadata"] is None
